#!BPY
 
"""
Name: 'Skyrim nif (.nif)...'
Blender: 249
Group: 'Import'
Tooltip: 'Skyrim NIF file.'
"""

__author__= "himika"
__version__= "0.1"

__bpydoc__= """\
This script imports a NIF file to Blender.

Usage:
Run this script from "File->Import" menu and then load the desired NIF file
"""

# ***** BEGIN BSD LICENSE BLOCK *****
#
# BSD License
# 
# Copyright (c) 2012 himika
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The name of the NIF File Format Library and Tools project may not be
#    used to endorse or promote products derived from this software
#    without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# ***** END BSD LICENSE BLOCK *****
# --------------------------------------------------------------------------

from struct import pack, unpack
import bpy
import Blender
from Blender import Mesh, Armature, Draw, Window, Material, Texture
from Blender.Mathutils import Vector
import BPyMessages

class NifFormat():
	class SkinTransform():
		def __init__(self):
			self.rotation = []
			self.translation = []
			self.scale = 0
	
	class SkinData():
		def __init__(self):
			self.skin_transform = NifFormat.SkinTransform()
			self.bounding_sphere_offset = [0,0,0]
			self.bounding_sphere_radius = 0
			self.num_vertices = 0
			self.vertex_weights = []
	
	class Header():
		def __init__(self):
			self.length = 0
			self.header_string = ""			# string
			self.copyright = ""			# string
			self.version = [0,0,0,0]		# byte byte byte byte
			self.endian = 0				# byte
			self.user_version = 0			# int
			self.num_blocks = 0			# int
			self.user_verson2 = 0			# int
			self.unknown_int_3 = 0			# int
			self.export_info = ['','','']		# short_string*3
			self.num_block_types = 0		# ushort
			self.block_types = []			# (uint + string)*
			self.block_type_index = []		# ushort*
			self.block_size = []			# uint*
			self.num_strings = 0			# uint
			self.max_string_length = 0		# uint
			self.strings = []			# (uint + string)*
			self.unknown_int_2 = 0			# uint
		
		def read(self, f):
			start_pos = f.tell()
			
			data = f.read(1)
			while data != "\x0A":
				self.header_string += data
				data = f.read(1)
			
			data = f.read(4)
			vers = unpack('<4B', data)
			self.version = [vers[3], vers[2], vers[1], vers[0]]
			if ("%d.%d.%d.%d" % vers) != "7.0.2.20":
				raise(ValueError, 'version error')
			
			data = f.read(1)
			self.endian, = unpack('<B', data)
			
			data = f.read(4*3)
			self.user_version, self.num_blocks, self.user_version2 = unpack('<III', data)
			if (self.user_version != 12 or self.user_version2 != 83):
				raise(ValueError, 'version error')
			
			data = f.read(1)
			n, = unpack('<B', data)
			self.export_info[0] = f.read(n)
			data = f.read(1)
			n, = unpack('<B', data)
			self.export_info[1] = f.read(n)
			data = f.read(1)
			n, = unpack('<B', data)
			self.export_info[2] = f.read(n)
			
			data = f.read(2)
			self.num_block_types, = unpack('<H', data)
			
			self.block_types = []
			for i in range(self.num_block_types):
				data = f.read(4)
				n, = unpack('<I', data)
				block_name = f.read(n)
				self.block_types.append(block_name)
			
			self.block_type_index = []
			for i in range(self.num_blocks):
				data = f.read(2)
				idx, = unpack('<H', data)
				self.block_type_index.append(idx)
			
			self.block_size = []
			for i in range(self.num_blocks):
				data = f.read(4)
				n, = unpack('<I', data)
				self.block_size.append(n)
			
			data = f.read(4*2)
			self.num_strings, self.max_string_length = unpack('<II', data)
			
			self.strings = []
			for i in range(self.num_strings):
				data = f.read(4)
				n, = unpack('<I', data)
				s = f.read(n)
				self.strings.append(s)
			
			data = f.read(4)
			self.unknown_int_2, = unpack('<I', data)
			
			end_pos = f.tell()
			self.length = end_pos - start_pos
		
		def write(self, f):
			f.write(self.header_string + "\x0A")
			f.write(pack('<BBBB', self.version[3], self.version[2], self.version[1], self.version[0]))
			f.write(pack('<B', self.endian))
			f.write(pack('<III', self.user_version, self.num_blocks, self.user_version2))
			f.write(pack('<B', len(self.export_info[0])) + self.export_info[0])
			f.write(pack('<B', len(self.export_info[1])) + self.export_info[1])
			f.write(pack('<B', len(self.export_info[2])) + self.export_info[2])
			f.write(pack('<H', self.num_block_types))
			for i in range(self.num_block_types):
				f.write(pack('<I', len(self.block_types[i])) + self.block_types[i])
			for i in range(self.num_blocks):
				f.write(pack('<H', self.block_type_index[i]))
			for i in range(self.num_blocks):
				f.write(pack('<I', self.block_size[i]))
			f.write(pack('<II', self.num_strings, self.max_string_length))
			for i in range(self.num_strings):
				f.write(pack('<I', len(self.strings[i])) + self.strings[i])
			f.write(pack('<I', self.unknown_int_2))
	
	class Footer():
		def __init__(self):
			self.length = 0
			self.num_roots = 0
			self.roots = []
		
		def read(self, f):
			data = f.read(4)
			self.num_roots, = unpack('<I', data)
			
			self.roots = []
			for i in range(self.num_roots):
				data = f.read(4)
				root, = unpack('<I', data)
				self.roots.append(root)
			
			self.length = 4 + 4*self.num_roots
		
		def write(self, f):
			f.write(pack('<I', len(self.roots)))
			for i in range(self.num_roots):
				f.write(pack('<I', self.roots[i]))
	
	class NiObject():
#		def __init__(self, length=0, type=None):
		def __init__(self, length, type):
			self.length = length
			if type:
				self.type = type
			else:
				self.type = self.__class__.__name__
			self.raw = ""
			self.raw_pos = 0
		
		def unpack_raw_uint(self):
			uint, = unpack('<I', self.raw[self.raw_pos : self.raw_pos+4])
			self.raw_pos += 4
			return uint
		
		def unpack_raw_int(self):
			signed_int, = unpack('<i', self.raw[self.raw_pos : self.raw_pos+4])
			self.raw_pos += 4
			return signed_int
		
		def unpack_raw_ushort(self):
			ushort, = unpack('<H', self.raw[self.raw_pos : self.raw_pos+2])
			self.raw_pos += 2
			return ushort
		
		def unpack_raw_ubyte(self):
			ubyte, = unpack('<B', self.raw[self.raw_pos : self.raw_pos+1])
			self.raw_pos += 1
			return ubyte
		
		def unpack_raw_float(self):
			flo, = unpack('<f', self.raw[self.raw_pos : self.raw_pos+4])
			self.raw_pos += 4
			return flo
		
		def unpack_raw_color3(self):
			col4 = unpack('<fff', self.raw[self.raw_pos : self.raw_pos+4*3])
			self.raw_pos += 4*3
			return [c for c in col4]
		
		def unpack_raw_color4(self):
			col4 = unpack('<ffff', self.raw[self.raw_pos : self.raw_pos+4*4])
			self.raw_pos += 4*4
			return [c for c in col4]
		
		def unpack_raw_texcoord(self):
			u,v = unpack('<ff', self.raw[self.raw_pos : self.raw_pos+4*2])
			self.raw_pos += 4*2
			return [u,v]
		
		def unpack_raw_vector3(self):
			x,y,z = unpack('<fff', self.raw[self.raw_pos : self.raw_pos+4*3])
			self.raw_pos += 4*3
			return [x,y,z]
		
		def unpack_raw_vector4(self):
			x,y,z,a = unpack('<ffff', self.raw[self.raw_pos : self.raw_pos+4*4])
			self.raw_pos += 4*4
			return [x,y,z,a]
		
		def unpack_raw_triangle(self):
			a,b,c = unpack('<HHH', self.raw[self.raw_pos : self.raw_pos+2*3])
			self.raw_pos += 2*3
			return [a,b,c]
		
		def unpack_raw_matrix33(self):
			matrix33 = [ n for n in unpack('<9f', self.raw[self.raw_pos : self.raw_pos+4*3*3]) ]
			self.raw_pos += 4*3*3
			return matrix33
		
		def unpack_raw_skin_transform(self):
			st = NifFormat.SkinTransform()
			st.rotation = self.unpack_raw_matrix33()
			st.translation = self.unpack_raw_vector3()
			st.scale = self.unpack_raw_float()
			return st
		
		def unpack_raw_skin_data(self):
			sd = NifFormat.SkinData()
			sd.skin_transform = self.unpack_raw_skin_transform()
			sd.bounding_sphere_offset = self.unpack_raw_vector3()
			sd.bounding_sphere_radius = self.unpack_raw_float()
			sd.num_vertices = self.unpack_raw_ushort()
			sd.vertex_weights = []
			if sd.num_vertices:
				for i in range(sd.num_vertices):
					idx = self.unpack_raw_ushort()
					wgt = self.unpack_raw_float()
					sd.vertex_weights.append([idx,wgt])
			return sd
		
		def parse_read(self):
			return
		
		def parse_write(self):
			return
		
		def read(self, f):
			if self.length:
				self.raw = f.read(self.length)
				self.parse_read()
		
		def write(self, f):
			if self.length:
				self.parse_write()
				f.write(self.raw)
	
	class NiAVObject(NiObject):
		def parse_read(self):
			self.name = self.unpack_raw_uint()
			self.num_extra_data_list = self.unpack_raw_uint()
			self.extra_data_list = []
			if self.num_extra_data_list > 0:
				for i in range(self.num_extra_data_list):
					data = self.unpack_raw_float()
					self.extra_data_list.append(data)
			self.controller = self.unpack_raw_uint()
			self.flags = self.unpack_raw_ushort()
			self.unknown_short = self.unpack_raw_ushort()
			self.translation = self.unpack_raw_vector3()
			self.rotation = self.unpack_raw_matrix33()
			self.scale = self.unpack_raw_float()
			self.collision_object = self.unpack_raw_uint()
	
	class NiNode(NiAVObject):
		def parse_read(self):
			NifFormat.NiAVObject.parse_read(self)
			self.num_children = self.unpack_raw_uint()
			self.children = []
			if self.num_children > 0:
				for i in range(self.num_children):
					data = self.unpack_raw_uint()
					self.children.append(data)
			self.num_effects = self.unpack_raw_uint()
			self.effects = []
			if self.num_effects > 0:
				for i in range(self.num_effects):
					data = self.unpack_raw_uint()
					self.effects.append(data)
	class BSFadeNode(NiNode):
		dummy = 0
	
	class NiSkinInstance(NiObject):
		def parse_read(self):
			NifFormat.NiObject.parse_read(self)
			self.data = self.unpack_raw_uint()
			self.skin_partition = self.unpack_raw_uint()
			self.skeleton_root = self.unpack_raw_uint()
			self.num_bones = self.unpack_raw_uint()
			self.bones = []
			for i in range(self.num_bones):
				self.bones.append( self.unpack_raw_uint() )
	
	class BSDismemberSkinInstance(NiSkinInstance):
		def parse_read(self):
			NifFormat.NiSkinInstance.parse_read(self)
			self.num_partitions = self.unpack_raw_uint()
			self.partitions = []
			for i in range(self.num_partitions):
				part_flag = self.unpack_raw_ushort()
				body_part = self.unpack_raw_ushort()
				self.partitions.append([part_flag, body_part])
	
	class NiSkinData(NiObject):
		def parse_read(self):
			NifFormat.NiObject.parse_read(self)
			self.skin_transform = self.unpack_raw_skin_transform()
			self.num_bones = self.unpack_raw_uint()
			self.has_vertex_weights = self.unpack_raw_ubyte()
			self.bone_list = []
			if self.has_vertex_weights:
				for i in range(self.num_bones):
					bone_list = self.unpack_raw_skin_data()
					self.bone_list.append(bone_list)
	
	class NiTriShape(NiAVObject):
		def parse_read(self):
			NifFormat.NiAVObject.parse_read(self)
			self.data = self.unpack_raw_uint()
			self.skin_instance = self.unpack_raw_uint()
			self.num_materials = self.unpack_raw_uint()
			self.material_name = []
			for i in range(self.num_materials):
				self.material_name.append( self.unpack_raw_uint() )
			self.material_extra_data = []
			for i in range(self.num_materials):
				self.material_extra_data.append( self.unpack_raw_int() )
			self.active_material = self.unpack_raw_int()
			self.dirty_flag = self.unpack_raw_ubyte()
			self.properties = [ self.unpack_raw_uint(), self.unpack_raw_uint() ]
			return
	
	class NiTriShapeData(NiObject):
		def parse_read(self):
			self.unknown_int = self.unpack_raw_uint()
			self.num_vertices = self.unpack_raw_ushort()
			self.keep_flags = self.unpack_raw_ubyte()
			self.compress_flags = self.unpack_raw_ubyte()
			self.has_vertices = self.unpack_raw_ubyte()
			self.vertices = []
			if self.has_vertices:
				for i in range(self.num_vertices):
					self.vertices.append( self.unpack_raw_vector3() )
			self.num_uv_sets = self.unpack_raw_ushort()
			self.unknown_int_2 = self.unpack_raw_uint()
			self.has_normals = self.unpack_raw_ubyte()
			self.normals = []
			self.tangents = []
			self.binormals = []
			if self.has_normals:
				for i in range(self.num_vertices):
					self.normals.append( self.unpack_raw_vector3() )
				for i in range(self.num_vertices):
					self.tangents.append( self.unpack_raw_vector3() )
				for i in range(self.num_vertices):
					self.binormals.append( self.unpack_raw_vector3() )
			self.center = self.unpack_raw_vector3()
			self.radius = self.unpack_raw_float()
			self.has_vertex_colors = self.unpack_raw_ubyte()
			self.vertex_colors = []
			if self.has_vertex_colors:
				for i in range(self.num_vertices):
					self.vertex_colors.append( self.unpack_raw_color4() )
			self.uv_sets = []
			if (self.num_uv_sets & 1):
				for i in range(self.num_vertices):
					self.uv_sets.append( self.unpack_raw_texcoord() )
			self.consistency_flags = self.unpack_raw_ushort()
			self.additional_data = self.unpack_raw_uint()
			self.num_triangles = self.unpack_raw_ushort()
			self.num_triangle_points = self.unpack_raw_uint()
			self.has_triangles = self.unpack_raw_ubyte()
			self.triangles = []
			if self.has_triangles:
				for i in range(self.num_triangles):
					self.triangles.append( self.unpack_raw_triangle() )
			self.num_match_groups = self.unpack_raw_ushort()
			self.match_groups = []
			if self.num_match_groups:
				for i in range(self.num_match_groups):
					num_vertices = self.unpack_raw_ushort()
					vertex_indices = self.unpack_raw_ushort()
					self.match_groups.append( [num_vertices, vertex_indices] )
			return
		
		def parse_write(self):
			if self.has_vertices:
				data = ""
				for i in range(self.num_vertices):
					v = self.vertices[i]
					data += pack('<fff', v[0], v[1], v[2])
				self.raw = self.raw[:9] + data + self.raw[9+4*3*self.num_vertices:]
	
	
	class BSLightingShaderProperty(NiObject):
		SLSF2_ZBuffer_Write = 1
		SLSF2_Double_Sided = 0x10
		SLSF2_EnvMap_Light_Fade = 0x8000
		
		def parse_read(self):
			self.skyrim_shader_type = self.unpack_raw_uint()
			self.name = self.unpack_raw_uint()
			self.num_extra_data_list = self.unpack_raw_uint()
			self.extra_data_list = []
			for i in range(self.num_extra_data_list):
				self.extra_data_list.append( self.unpack_raw_uint() )
			self.controller = self.unpack_raw_uint()
			self.shader_flags_1 = self.unpack_raw_uint()
			self.shader_flags_2 = self.unpack_raw_uint()
			self.uv_offset = self.unpack_raw_texcoord()
			self.uv_scale = self.unpack_raw_texcoord()
			self.texture_set = self.unpack_raw_uint()
			self.emissive_color = self.unpack_raw_color3()
			self.emissive_multiple = self.unpack_raw_float()
			self.texture_clamp_mode = self.unpack_raw_uint()
			self.alpha = self.unpack_raw_float()
			self.unknown_float = self.unpack_raw_float()
			self.specular_power_glossiness = self.unpack_raw_float()
			self.specular_color = self.unpack_raw_color3()
			self.specular_strength = self.unpack_raw_float()
			self.lighting_effect_1 = self.unpack_raw_float()
			self.lighting_effect_2 = self.unpack_raw_float()
			if self.skyrim_shader_type == 1:
				self.environment_map_scale = self.unpack_raw_float()
			if self.skyrim_shader_type == 5:
				self.skin_tint_color = self.unpack_raw_color3()
			if self.skyrim_shader_type == 6:
				self.hair_tint_color = self.unpack_raw_color3()
			if self.skyrim_shader_type == 7:
				self.max_passes = self.unpack_raw_()
				self.scale = self.unpack_raw_()
			if self.skyrim_shader_type == 11:
				self.parallax_inner_layer_thickness = self.unpack_raw_float()
				self.parallax_refraction_scale = self.unpack_raw_float()
				self.parallax_inner_layer_texture_scale = self.unpack_raw_texcoord()
				self.parallax_envmap_strength = self.unpack_raw_float()
			if self.skyrim_shader_type == 14:
				self.sparkle_paramaters = self.unpack_raw_vector4()
			if self.skyrim_shader_type == 16:
				self.eye_cubemap_scale = self.unpack_raw_float()
				self.left_eye_reflection_center = self.unpack_raw_vector3()
				self.right_eye_reflection_center = self.unpack_raw_vector3()
	
	
	class NifBlock(object):
		def __new__(mcls, block_size, block_type):
			try:
				cls = eval('NifFormat.' + block_type)
				if not issubclass(cls, NifFormat.NiObject):
					cls = NifFormat.NiObject
			except AttributeError:
				cls = NifFormat.NiObject
			return cls(block_size, block_type)
	
	
	def __init__(self):
		self.header = self.Header()
		self.blocks = []
		self.footer = self.Footer()
	
	
	def read(self, f):
		self.header.read(f)
		
		for i in range(self.header.num_blocks):
			block_size = self.header.block_size[i]
			block_type = self.header.block_types[ self.header.block_type_index[i] ]
			block = NifFormat.NifBlock(block_size, block_type)
			block.read(f)
			self.blocks.append(block)
		
		self.footer.read(f)
	
	def write(self, f):
		self.header.write(f)
		
		for i in range(self.header.num_blocks):
			self.blocks[i].write(f)
		
		self.footer.write(f)



def bonedict_has_key(dict, key):
	try:
		dict.keys().index(key)
		return True
	except ValueError:
		return False



class NifImporter:
	def __init__(self, filename):
		self.nif = NifFormat()
		self.scene = bpy.data.scenes.active
		self.objects = {}
		self.filename = filename
		
		self.scalefactor = 1
		self.weld_seams = False
	
	
	def read(self):
		f = open(self.filename, 'rb')
		self.nif = NifFormat()
		self.nif.read(f)
		f.close()
	
	
	def parse(self):
		for i in self.nif.footer.roots:
			self._walker(i)
		print self.objects
		
		self.scene.update(1)
		
		for i in self.objects.keys():
			ob = self.objects[i]
			block = self.nif.blocks[i]
			if block.type == 'NiTriShape':
				if (block.skin_instance != 0xffffffff):
					niSkinInstance = self.nif.blocks[block.skin_instance]
					ob_armature = self.objects[niSkinInstance.skeleton_root]
					ob_armature.makeParent([ob])
		
		return (ob for ob in self.objects.values())
	
	
	def _walker(self, i):
		block = self.nif.blocks[i]
		if block.type == 'NiTriShape':
			ob = self.parse_NiTriShape(i)
		elif block.type == 'NiNode' or block.type == 'BSFadeNode':
			for j in block.children:
				self._walker(j)
	
	
	def parse_NiTriShape(self, i):
		block = self.nif.blocks[i]
		name = self.nif.header.strings[block.name]
		scale = block.scale
		
		mesh = self.parse_NiTriShapeData(block.data, name)
		ob = self.scene.objects.new(mesh, name)
		ob.loc = ( n * self.scalefactor for n in block.translation)
		ob.size = (scale, scale, scale)
		self.objects[i] = ob
		
		if (block.skin_instance != 0xffffffff):
			ob_armature = self.parse_NiSkinInstance(mesh, block.skin_instance)
		
		for j in block.properties:
			if (j != 0xffffffff):
				prop = self.nif.blocks[j]
				if prop.type == 'BSLightingShaderProperty':
					self.parse_BSLightingShaderProperty(ob, j)
		
		return ob
	
	
	def parse_NiTriShapeData(self, i, name):
		mesh = Mesh.New(name)
		mesh.mode = 0
		block = self.nif.blocks[i]
		
		# verts
		if block.has_normals and self.weld_seams:
			v_dict = {}
			v_map  = []
			verts_list = []
			for i in range(block.num_vertices):
				v = block.vertices[i]
				n = block.normals[i]
				key = pack('<ffffff', v[0], v[1], v[2], n[0], n[1], n[2])
				if v_dict.has_key(key):
					v_map.append(v_dict[key])
				else:
					j = len(verts_list)
					v_dict[key] = j
					v_map.append(j)
					v = [ f * self.scalefactor for f in v ]
					verts_list.append( tuple(v) )
			del v_dict
		else:
			v_map = range(0, block.num_vertices)
			verts_list = [(v[0]*self.scalefactor, v[1]*self.scalefactor, v[2]*self.scalefactor) for v in block.vertices]
		
		mesh.verts.extend(verts_list)
		del verts_list
		
		# face
		if block.has_triangles and block.num_triangles > 0:
			face_list = [(v_map[f[0]], v_map[f[1]], v_map[f[2]]) for f in block.triangles]
			mesh.faces.extend(face_list)
		del face_list
		
		# UV
		if (block.num_uv_sets & 1):
			mesh.faceUV = True
			for i, tri in enumerate(block.triangles):
				for j, k in enumerate(tri):
					coord = block.uv_sets[k]
					mesh.faces[i].uv[j].x = coord[0]
					mesh.faces[i].uv[j].y = 1-coord[1]
		
		
		# Vertex Color
		if block.has_vertex_colors:
			mesh.vertexColors = 1
			for i, tri in enumerate(block.triangles):
				for j, k in enumerate(tri):
					vcol = block.vertex_colors[ v_map[k] ]
					mesh.faces[i].col[j].r = int(vcol[0] * 255)
					mesh.faces[i].col[j].g = int(vcol[1] * 255)
					mesh.faces[i].col[j].b = int(vcol[2] * 255)
					mesh.faces[i].col[j].a = int(vcol[3] * 255)
		
		# Normal
		if block.has_normals:
			for i in range(block.num_vertices):
				idx = v_map[i]
				normal = block.normals[i]
				mesh.verts[idx].no = Vector(normal[0], normal[1], normal[2])
			for f in mesh.faces:
				f.smooth = 1
		
		self.v_map = v_map
		return mesh
	
	
	def parse_NiSkinInstance(self, mesh, i):
		block = self.nif.blocks[i]
		skin_data = self.nif.blocks[ block.data ]
		list_vgroup = []
		replace = Blender.Mesh.AssignModes.REPLACE
		v_map = self.v_map
		
		niRoot = self.nif.blocks[ block.skeleton_root ]
		skeleton_name = self.nif.header.strings[ niRoot.name ]
		
		if self.objects.has_key( block.skeleton_root ):
			ob = self.objects[block.skeleton_root]
			arm = ob.getData()
		else:
			arm = Armature.New()
			arm.drawType = Armature.STICK
			arm.drawAxes = True
			arm.envelopes = False
			ob = self.scene.objects.new(arm, skeleton_name)
			self.objects[block.skeleton_root] = ob
		
		arm.makeEditable()
		for i in block.bones:
			group_name = self.nif.header.strings[ self.nif.blocks[i].name ]
			list_vgroup.append(group_name)
			mesh.addVertGroup(group_name)
			
			if not bonedict_has_key(arm.bones, group_name):
				niBone = self.nif.blocks[i]
				eb = Armature.Editbone()
				eb.roll = 0
				eb.head = Vector(niBone.translation[0], niBone.translation[1], niBone.translation[2]) * self.scalefactor
				eb.tail = Vector(niBone.translation[0]+5, niBone.translation[1], niBone.translation[2]) * self.scalefactor
				arm.bones[group_name] = eb
		arm.update()
		
		for i in range(skin_data.num_bones):
			bone_list = skin_data.bone_list[i]
			group_name = list_vgroup[i]
			for vertex_weight in bone_list.vertex_weights:
				idx = vertex_weight[0]
				wgt = vertex_weight[1]
				mesh.assignVertsToGroup(group_name, [v_map[idx]], wgt, replace)
		
		return ob
	
	def parse_BSLightingShaderProperty(self, ob, i):
		block = self.nif.blocks[i]
		mesh = ob.getData(mesh=True)
		name = ob.getName()
		
		if (block.shader_flags_2 & NifFormat.BSLightingShaderProperty.SLSF2_Double_Sided):
			mesh.mode = (mesh.mode & ~Mesh.ModesTWOSIDED) | Mesh.Modes.TWOSIDED
		if (block.texture_set != 0xffffffff):
			tex = Texture.New('Tex.'+name)
			tex.setType('Image')
			texco = Blender.Texture.TexCo.UV
			mapto = Blender.Texture.MapTo.COL | Blender.Texture.MapTo.EMIT
			mat = Material.New('Mat.'+name)
			mat.setTexture(0, tex, texco, mapto)
			mesh.materials += [mat]


def loadNif(filename):
	if BPyMessages.Error_NoFile(filename):
		return
	
	print('\nImporting nif: "%s"' % (Blender.sys.expandpath(filename)))
	importer = NifImporter(filename)
	importer.read()
	importer.parse()


# main
if __name__=='__main__':
	Blender.Window.FileSelector(loadNif, 'Import Nif', '*.nif')

